/*
  Copyright(2006) CNRS
 
  Authors: Florent Lamiraux (CNRS) Mathieu Poirier (CNRS)

*/

/*****************************************
 INCLUDES
*******************************************/

#include <iostream>
#include <fstream>

#include "KineoModel/kppSolidComponentRef.h"

#include "hppWalkPlanner/hppWalkPlanner.h"
#include "hppWalkPlanner/hppWalkPatGen.h"
#include "hrp2_14/hrp2_14.h"

// Select level of verbosity at configuration.
#if DEBUG==3
#define ODEBUG3(x) std::cout << "ChppWalkPlanner:" << x << std::endl
#define ODEBUG2(x) std::cout << "ChppWalkPlanner:" << x << std::endl
#define ODEBUG1(x) std::cerr << "ChppWalkPlanner:" << x << std::endl
#elif DEBUG==2
#define ODEBUG3(x)
#define ODEBUG2(x) std::cout << "ChppWalkPlanner:" << x << std::endl
#define ODEBUG1(x) std::cerr << "ChppWalkPlanner:" << x << std::endl
#elif DEBUG==1
#define ODEBUG3(x)
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "ChppWalkPlanner:" << x << std::endl
#else
#define ODEBUG3(x)
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

// ==========================================================================


ChppWalkPlanner::ChppWalkPlanner(double inStepLength, double inStepSpace, 
				 double inFootSecurityDistance,
				 double inBoxLateralTolerance, const char *model) : 
  ChppWalkFootPlanner(inStepLength, inStepSpace, inFootSecurityDistance), 
  attHppPatGen(inStepLength, inStepSpace, model),
  attBoxLateralTolerance(inBoxLateralTolerance)
{
  attHRP2Path.clear() ;
  attHRP2TimeRef.clear() ;
  attHRP2ZMPRef.clear() ;
  attHRP2WaistPosition.clear() ;
}

// ==========================================================================


ChppWalkPlanner::~ChppWalkPlanner(){

 
  attHRP2Path.clear() ;
  attHRP2TimeRef.clear() ;
  attHRP2ZMPRef.clear() ;
  attHRP2WaistPosition.clear() ;

}

// ==========================================================================

ktStatus ChppWalkPlanner::initializeProblem()
{

  double xMin=0, yMin=0, zMin=0, xMax=0, yMax=0, zMax=0;
  unsigned int nbRobots = getNbHppProblems();

  

  if (nbRobots == 0) {
    ODEBUG1(":initializeProblem: there is no robot.");
    return KD_ERROR;
  }

  // Assume humanoid robot is at rank 0 in problem vector.
  ChppDeviceShPtr humanoidRobot = KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem(0));
   
  // Compute bouding box of humanoid robot.
  if (humanoidRobot->axisAlignedBoundingBox(xMin, yMin, zMin, xMax, yMax, zMax) != KD_OK) {
    ODEBUG1(":initializeProblem: robot seems to have no geometry");
    return KD_ERROR;
  }
  const ChppDeviceShPtr& robotBox = createRobotBox(xMin, yMin-attBoxLateralTolerance, -0.01, 
						   xMax, yMax+attBoxLateralTolerance, zMax);
  if (!robotBox) {
    ODEBUG1(":initializeProblem: failed to create robot bounding box.");    
    return KD_ERROR;
  }

  //make the joint invisible by default
  robotBox->isVisible(false) ;
 
  //the body of the bouding box transparent
  robotBox->solidComponentRef(0)->referencedSolidComponent()->isTransparent(true) ;

  // Set same default penetration for box and humanoid
  double boxPenetration = penetration(0);
  // Register bounding box in problem vector.
  addHppProblemAtBeginning(robotBox, boxPenetration);

  // desactived HRP2 and box collision 
  robotBox->ignoreDeviceForCollision(humanoidRobot);
  humanoidRobot->ignoreDeviceForCollision(robotBox);
  
  if (ChppWalkFootPlanner::initializeProblem() != KD_OK) {
    return KD_ERROR;
  }
  return KD_OK;
}

// ==========================================================================

ktStatus ChppWalkPlanner::initConfig(double inX, double inY, double inTheta)
{
  if (ChppWalkFootPlanner::initConfig(inX, inY, inTheta) != KD_OK) {
    return KD_ERROR;
  }
  
  if (getNbHppProblems() < 2) {
    ODEBUG1(":initConfig: not enought robot has been defined : " << getNbHppProblems());
    return KD_ERROR;
  }

  /*
    Build configuration for HRP2.
  */

  const ChppDeviceShPtr HRP2Device = KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem(1));

  if( !HRP2Device){
    ODEBUG1(":initConfig: HRP2 is empty") ;
  }

  CkwsConfigShPtr  initConfHRP2 =  CkwsConfig::create(HRP2Device);
  initConfHRP2->dofValue(0, inX);
  initConfHRP2->dofValue(1, inY);
  initConfHRP2->dofValue(2, 0);
  initConfHRP2->dofValue(3, 0);
  initConfHRP2->dofValue(4, 0);
  initConfHRP2->dofValue(5, inTheta);
  
  if (!initConfHRP2) {
    ODEBUG1(":initConfig: could not create the init configuration HRP2.");
    return KD_ERROR;
  }
    
  if (initConfIthProblem(1, initConfHRP2) == KD_ERROR) {
    ODEBUG1(":initConfig could not set initial configuration of robot HRP2.");
    return KD_ERROR;
  }

  return KD_OK;
}

// ==========================================================================

ktStatus ChppWalkPlanner::goalConfig(double inX, double inY, double inTheta)
{
  if (ChppWalkFootPlanner::goalConfig(inX, inY, inTheta) != KD_OK) {
    return KD_ERROR;
  }
  
  if (getNbHppProblems() < 2) {
    ODEBUG1(":goalConfig - not enought robot has been defined : " << getNbHppProblems());
    return KD_ERROR;
  }

  /*
    Build configuration for HRP2.
  */

  const ChppDeviceShPtr HRP2Device = KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem(1));
  
  if( !HRP2Device){
    ODEBUG1(":goalConfig: HRP2 is empty") ;
  }

  CkwsConfigShPtr  goalConfHRP2 =  CkwsConfig::create(HRP2Device);
  goalConfHRP2->dofValue(0, inX);
  goalConfHRP2->dofValue(1, inY);
  goalConfHRP2->dofValue(2, 0);
  goalConfHRP2->dofValue(3, 0);
  goalConfHRP2->dofValue(4, 0);
  goalConfHRP2->dofValue(5, inTheta);
    
  if (!goalConfHRP2) {
    ODEBUG1(":goalConfig: could not create the goal configuration HRP2.");
    return KD_ERROR;
  }
    
  if (goalConfIthProblem(1, goalConfHRP2) == KD_ERROR) {
    ODEBUG1(":goalConfig could not set initial configuration of robot HRP2.");
    return KD_ERROR;
  }

  return KD_OK;
}

// ==========================================================================


ktStatus ChppWalkPlanner::initOpenHRPPosturePatGenRad(double initPos[40]) 
{
  return attHppPatGen.initOpenHRPPosturePatGenRad(initPos);
}

// ==========================================================================

ktStatus ChppWalkPlanner::solve()
{
  ktStatus status ;
  status = ChppWalkFootPlanner::solve();
  if (status == KD_ERROR) {
    return status;
  }
  
  ODEBUG2(":solve:______________ PHASE 3: Humanoid trajectory________" ) ;
  
  status = computeHumanoidTrajectory();

  return status ;
}


// ==========================================================================

ktStatus ChppWalkPlanner::computeHumanoidTrajectory()
{
  CkwsPathShPtr pathHrp2 = CkwsPath::create(robotIthProblem(1)) ;
  attHppPatGen.reset();
  resetHumanoidTrajectory();
  ODEBUG2(" clear all vector ... " );

  ODEBUG2(" sequenceOfStepSequences size = " <<  sequenceOfStepSequences().size() );

  
  for (unsigned int iPath=0; iPath < sequenceOfStepSequences().size(); iPath++) {

    if(attHppPatGen.computeAllQFromAStepSequence(KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem(1)),  
						 pathHrp2, attHRP2Path, 
						 attHRP2TimeRef, attHRP2ZMPRef, attHRP2WaistPosition,
						 sequenceOfStepSequences()[iPath])) {
      ODEBUG1(":computeHumanoidTrajectory: detected collision in walking motion.");
      addPath(1,pathHrp2);
      return KD_ERROR;
    }
  }
  // put the REAL Config in init and goal configs of HRP2 
  initConfIthProblem(1,pathHrp2->configAtStart()); 
  goalConfIthProblem(1,pathHrp2->configAtEnd()) ;
 
  ODEBUG2(":computeHumanoidTrajectory: number of DP in path : " << pathHrp2->countDirectPaths() );
  addPath(1,pathHrp2);

  return KD_OK;
}

// ==========================================================================

ktStatus ChppWalkPlanner::solution( vector<CkwsConfigShPtr> &outHRP2Path, vector<double> &outHRP2TimeRef, 
				    vector<vector<double> > &outHRP2ZMPRef, vector<vector<double> > &outHRP2WaistPosition ) {

  if(attHRP2Path.empty()) {
    ODEBUG1(" ERROR : ChppWalkPlanner::solution - attHRP2Path empty ") ;
    return KD_ERROR ;
  }
  else outHRP2Path = attHRP2Path ;

  if(attHRP2TimeRef.empty()) {
    ODEBUG1(" ERROR : ChppWalkPlanner::solution - attHRP2TimeRef empty ") ;
    return KD_ERROR ;
  }
  else outHRP2TimeRef = attHRP2TimeRef ;

  if(attHRP2ZMPRef.empty()) {
    ODEBUG1(" ERROR : ChppWalkPlanner::solution - attHRP2ZMPRef empty ") ;
    return KD_ERROR ;
  }
  else outHRP2ZMPRef = attHRP2ZMPRef  ;

  if(attHRP2WaistPosition.empty()) {
    ODEBUG1(" ERROR : ChppWalkPlanner::solution - attHRP2WaistPosition empty ") ;
    return KD_ERROR ;
  }
  else outHRP2WaistPosition = attHRP2WaistPosition ;

   return KD_OK ;

}

